from typing import AsyncGenerator


class HTTPRequest:
    def __init__(
        self, method: str, uri: str, version: str, headers: dict[str, str], body: str
    ):
        self.method = method
        self.uri = uri
        self.version = version
        self.headers = headers
        self._body = body

    def __str__(self) -> str:
        return self.to_bytes().decode("utf-8")

    @property
    def body(self):
        return self._body

    @body.setter
    def set_body(self, v):
        self.body = v
        self.headers["Content-Length"] = len(v)

    def get_cookies(self) -> dict[str, str]:
        if "Cookie" not in self.headers:
            return dict()
        else:
            return dict([i.split("=", 1) for i in self.headers["Cookie"].split("; ")])

    def to_bytes(self):
        raw = (
            f"{self.method} {self.uri} {self.version}\r\n"
            + "\r\n".join([f"{k}: {v}" for k, v in self.headers.items()])
            + "\r\n\r\n"
            + self._body
        )
        return raw.encode("utf-8")

    @staticmethod
    async def parse(data: AsyncGenerator[bytes, None]) -> "HTTPRequest":
        request_data, body, body_length, headers = None, None, 0, None

        raw = ""
        async for chunk in data:
            raw += chunk.decode("utf-8")
            if request_data is None and "\r\n" in raw:
                request_data, raw = raw.split("\r\n", 1)
            if "\r\n\r\n" in raw:
                headers, raw = raw.split("\r\n\r\n", 1)
                headers = dict([i.split(": ") for i in headers.split("\r\n")])
                body_length = int(headers.get("Content-Length", 0))
            if len(raw) >= body_length:
                body = raw
                break

        method, uri, version = request_data.split(" ")
        return HTTPRequest(
            method=method, uri=uri, version=version, headers=headers, body=body
        )



import asyncio

from .classes.config import Config, ServerConfig
from .classes.http_request import HTTPRequest
from .classes.page_loader import PageLoader
from .cookie import create_cookie, get_cookie_data


def redirect_answer(username: str, config: Config):
    return (
        b"HTTP/1.1 302 Found\r\nLocation: /\r\nSet-Cookie: totp_auth="
        + create_cookie(username, config).encode("utf-8")
        + b"\r\n\r\n"
    )


async def receive_request(reader):
    while True:
        yield await reader.read(4096)


async def forward_to_client(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
    while True:
        data = await reader.read(4096)
        if not data:
            break
        writer.write(data)
        await writer.drain()


async def forward_from_client(
    reader: asyncio.StreamReader, writer: asyncio.StreamWriter, config: ServerConfig
):
    if len(config.headers_rewrite) > 0:
        while True:
            request = await HTTPRequest.parse(receive_request(reader))
            for h, v in config.headers_rewrite.items():
                request.headers[h] = v
            writer.write(request.to_bytes())
    else:
        await forward_to_client(reader, writer)


async def proxy_request(
    request: HTTPRequest,
    local_reader: asyncio.StreamReader,
    local_writer: asyncio.StreamWriter,
    config: ServerConfig,
):
    remote_reader, remote_writer = await asyncio.open_connection(*config.redirect_data)
    for h, v in config.headers_rewrite.items():
        request.headers[h] = v
    remote_writer.write(request.to_bytes())

    await asyncio.gather(
        forward_from_client(local_reader, remote_writer, config),
        forward_to_client(remote_reader, local_writer),
    )


async def handle_client(
    local_reader: asyncio.StreamReader,
    local_writer: asyncio.StreamWriter,
    server_config: ServerConfig,
    config: Config,
    page_loader: PageLoader,
):
    request = await HTTPRequest.parse(receive_request(local_reader))
    cookies = request.get_cookies()

    if get_cookie_data(cookies.get("totp_auth", ""), config) in server_config.users:
        await proxy_request(request, local_reader, local_writer, server_config)
    else:
        if request.method == "POST":
            form_data = dict([i.split("=", 1) for i in request.body.split("&")])
            username, totp = form_data.get("username").lower(), form_data.get("totp")

            correct = False
            if username in server_config.users:
                if server_config.users[username].verify(totp):
                    correct = True

            if correct:
                local_writer.write(redirect_answer(username, config))
            else:
                local_writer.write(page_loader.render_response("Incorrect data"))
        else:
            local_writer.write(page_loader.render_response())

    local_writer.close()


def handle_client_decorator(config, server_config, page_loader):
    async def handler(reader, writer):
        return await handle_client(reader, writer, server_config, config, page_loader)

    return handler


async def start_server_async(config: Config):
    page_loader = PageLoader("./totp_auth/login.html")

    servers_data = []
    for i in config.list_servers():
        await asyncio.start_server(
            handle_client_decorator(config, i, page_loader), *i.listen_data
        )

        servers_data.append(f"{i.listen} -> {i.redirect}")

    print("Server started!\n" + "\n".join(servers_data))
    await asyncio.Future()


def start_server(config: Config):
    asyncio.run(start_server_async(config))



from configparser import ConfigParser, DuplicateSectionError
from contextlib import suppress
from dataclasses import dataclass
import secrets

from pyotp import TOTP


@dataclass
class ServerConfig:
    name: str
    listen: str
    redirect: str
    headers_rewrite: dict[str, str]
    users: dict[str, TOTP]

    @property
    def listen_data(self) -> tuple[str, int]:
        if ":" in self.listen:
            raw = self.listen.split(":")
            return raw[0], int(raw[1])
        else:
            return self.listen, 80

    @property
    def redirect_data(self) -> tuple[str, int]:
        if ":" in self.redirect:
            raw = self.redirect.split(":")
            return raw[0], int(raw[1])
        else:
            return self.redirect, 80


class Config:
    def __init__(self, filename: str) -> None:
        self.filename = filename
        self.config = ConfigParser()
        self.config.read(filename)

    def save(self) -> None:
        with open(self.filename, "w") as f:
            self.config.write(f)

    def reset_secret_token(self) -> None:
        self.config.set("DEFAULT", "secret_token", secrets.token_hex(16))

    def secret_token(self) -> str:
        return self.config.get("DEFAULT", "secret_token")

    def add_server(self, server: ServerConfig) -> None:
        with suppress(DuplicateSectionError):
            self.config.add_section(server.name)

        self.config.set(server.name, "listen", server.listen)
        self.config.set(server.name, "redirect", server.redirect)

        for h, v in server.headers_rewrite.items():
            self.config.set(server.name, f"header_{h.lower()}", v)

        for u, v in server.users.items():
            self.config.set(server.name, f"user_{u.lower()}", v.secret)

    def delete_server(self, name: str) -> None:
        self.config.remove_section(name)

    def get_server(self, name: str) -> ServerConfig:
        headers_rewrite = {}
        users = {}
        for j in self.config[name]:
            if j.startswith("header_"):
                headers_rewrite[j[7:].capitalize()] = self.config.get(name, j)
            elif j.startswith("user_"):
                users[j[5:]] = TOTP(self.config.get(name, j))

        return ServerConfig(
            name=name,
            listen=self.config.get(name, "listen"),
            redirect=self.config.get(name, "redirect"),
            headers_rewrite=headers_rewrite,
            users=users,
        )

    def list_servers(self) -> list[ServerConfig]:
        servers = []
        for i in self.config:
            if i == "DEFAULT":
                continue
            servers.append(self.get_server(i))

        return servers



import argparse
from contextlib import suppress

import pyotp

from .classes.config import Config, ServerConfig
from .server import start_server


def choose_server_cli(config: Config):
    servers = config.list_servers()
    for i, s in enumerate(servers):
        print(f"{i+1}. {s.name}")

    while True:
        s_id = input("Choose server id - ")
        if s_id.isnumeric():
            s_id = int(s_id)
            if s_id <= len(servers):
                break
    return servers[s_id - 1]


def user_cli(args: argparse.Namespace, config: Config):
    if args.user == "create":
        server = choose_server_cli(config)
        username = input("Choose username - ")
        code = pyotp.TOTP(pyotp.random_base32())
        print(f"Add code to your 2fa app - {code.secret}, 6 digit, SHA1")
        while True:
            test = input("Verify code from application - ")
            if code.verify(test):
                break
        server.users[username] = code
        config.add_server(server)
        print("Ready!")
    elif args.user == "delete":
        server = choose_server_cli(config)
        username = input("Choose username - ")
        if username not in server.users:
            print("User does not exists")
            return
        del server.users[username]
        config.add_server(server)
    else:
        print("Unknown command, use -h for help")
        return


def server_cli(args: argparse.Namespace, config: Config):
    if args.server == "ls":
        servers = config.list_servers()
        if len(servers) > 0:
            for i in servers:
                print(f"{i.name}: {i.listen} -> {i.redirect}")
        else:
            print(f"No servers configured")
    elif args.server == "create":
        name = input("Server name - ")
        listen = input("Listen address host:[port] - ")
        redirect = input("Redirect address host:[port] - ")
        config.add_server(ServerConfig(name, listen, redirect, {}, {}))
    elif args.server == "update":
        try:
            server = config.get_server(args.name)
        except KeyError:
            print(f"Server {args.name} does not exists")
            return

        try:
            k, v = args.data.split("=", 1)
        except ValueError:
            print("Expected key=value format")
            return

        config.delete_server(server.name)
        setattr(server, k, v)
        print("Updated!")
        config.add_server(server)
    else:
        print("Unknown command, use -h for help")
        return


def main():
    config = Config("config.ini")

    parser = argparse.ArgumentParser(description="Reverse proxy with TOTP auth")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")
    subparsers.add_parser("run", help="Start the server")

    subparsers.add_parser("reset_token", help="Reset secret token")

    parser_server = subparsers.add_parser("server", help="Set server address")
    server_subparsers = parser_server.add_subparsers(
        dest="server", help="Servers config"
    )

    server_subparsers.add_parser("ls", help="List all servers")
    server_subparsers.add_parser("create", help="New server")
    update_server = server_subparsers.add_parser("update", help="Update server")
    update_server.add_argument("name", type=str, help="Server name")
    update_server.add_argument("data", type=str, help="Key for change, key=value")

    parser_users = subparsers.add_parser("user", help="User login settings")
    users_subparsers = parser_users.add_subparsers(dest="user", help="Servers config")
    users_subparsers.add_parser("create", help="New server")
    users_subparsers.add_parser("delete", help="Delete user")

    args = parser.parse_args()

    with suppress(KeyboardInterrupt):
        if args.command == "run":
            start_server(config)
        elif args.command == "reset_token":
            config.reset_secret_token()
            print("Secret token has been reset")
        elif args.command == "server":
            server_cli(args, config)
        elif args.command == "user":
            user_cli(args, config)
        else:
            print("Unknown command, use -h for help")
            return

        config.save()


if __name__ == "__main__":
    main()



import base64
import hashlib
from typing import Optional

from .classes.config import Config


def _create_cookie_token(data: str, config: Config) -> str:
    return hashlib.sha256(f"{data}.{config.secret_token()}".encode("utf-8")).hexdigest()


def create_cookie(data: str, config: Config) -> str:
    token = _create_cookie_token(data, config)
    return base64.b64encode(data.encode("utf-8")).decode("utf-8") + "." + token


def get_cookie_data(cookie: str, config: Config) -> Optional[str]:
    raw = cookie.split(".", 1)
    if len(raw) != 2:
        return None
    raw_data, token = raw
    data = base64.b64decode(raw_data).decode("utf-8")
    if token != _create_cookie_token(data, config):
        return None
    else:
        return data



from typing import Optional


class PageLoader:
    def __init__(self, path: str):
        with open(path, "r") as f:
            self._page_raw = f.read()

    def render(self, error: Optional[str] = None):
        return self._page_raw.replace("{% ERROR %}", error or "")

    def render_response(self, error: Optional[str] = None):
        return b"HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n" + self.render(
            error
        ).encode("utf-8")
